import java.util.Map;
import java.util.Vector;
import java.util.StringTokenizer;
import se.datadosen.jalbum.*;
import se.datadosen.util.*;

/**
 * Title: Camera-specific helper code (relies upon the cameras.properties file in the JAlbum program directory) 
 * Copyright: Copyright (c) 2004, Olivier Biot
 * Author: Olivier Biot
 * Version: 1.2
 */
public class Camera {

	private Map meta = null;		// Image meta data
	private Map camera_info = null;		// Camera data
	private boolean allow_computation_with_focal_plane_resolution;	// Beware: set to FALSE if
									// your album contains
									// cropped images!


	/**
	 * Constructor.
	 * @param camera_info	The Map represented by the cameras.properties file.
	 */
	public Camera(Map camera_info)
	{
		this.camera_info = camera_info;
		this.allow_computation_with_focal_plane_resolution = false; // Default
	}


	/**
	 * Constructor.
	 * @param camera_info	The Map represented by the cameras.properties file.
	 * @param allow_computation_with_focal_plane_resolution Allow the 35mm focal equivalent to be computed based on the focal plane X/Y resolution (disable this if the photo was cropped)
	 */
	public Camera(Map camera_info, boolean allow_computation_with_focal_plane_resolution)
	{
		this.camera_info = camera_info;
		this.allow_computation_with_focal_plane_resolution = allow_computation_with_focal_plane_resolution;
	}


	//////////////////////////////////////////////////////////////////////////////////
	//
	// Helper methods
	//

	/**
	 * Strip non-numeric ending on strings and convert result to float.
	 * @param str The text string gontaining a floating point number.
	 * @return the parsed floating point value.
	 */
	public static double numericPart(String str) {
		int i;
		for (i=0; i < str.length(); i++) {
			char c = str.charAt(i);
			if (! Character.isDigit(c) && c != '.')
				break;
		}
		return Float.parseFloat(str.substring(0,i));
	}

	/**
	 * Strip the first character of a string.
	 * @param str The text string.
	 * @return the text string excluding the first character.
	 */
	public static String stripFirstChar(String str) {
		return str.substring(1, str.length());
	}

	/**
	 * Try to fetch metadata from a given key.
	 * @param meta	The meta-information map for the given file.
	 * @param key	The name of the key in the map.
	 * @return the value corresponding with the given key, or null.
	 */
	public String getValueForKey(Map meta, String key) {
		if (meta.containsKey(key)) {
			return meta.get(key).toString();
		}
		return null;
	}


	//////////////////////////////////////////////////////////////////////////////////
	//
	// 35mm focus computation code
	//

	// 35mm focus helper method.
	// Compute the image width from the "Exif Image Width" value.
	private double exifImageWidth(Map meta) {
		String width = getValueForKey(meta, "Exif Image Width");
		if (width == null || width.length() == 0)
			return 0.0; // We don't have this information
		StringTokenizer t = new StringTokenizer(width);
		// exif image width; just the number
		String imwStr = t.nextToken();
		// the width in pixels; integer
		return Double.parseDouble(imwStr);
	}

	// 35mm focus helper method.
	// Compute the image height from the "Exif Image Height" value.
	private double exifImageHeight(Map meta) {
		String height = getValueForKey(meta, "Exif Image Height");
		if (height == null || height.length() == 0)
			return 0.0; // We don't have this information
		StringTokenizer t = new StringTokenizer(height);
		// exif image width; just the number
		String imhStr = t.nextToken();
		// the width in pixels; integer
		return Double.parseDouble(imhStr);
	}

	// 35mm focus helper method.
	// Return the width of the CCD in mm.
	private double ccdWidth(Map meta) {
		String ccdxr = getValueForKey(meta, "Focal Plane X Resolution");
		if (ccdxr == null || ccdxr.length() == 0)
			return 0.0;
		// the resolution is a string, containing a fraction
		// [num]/[den] [unit]
		// for Canon camera's, [unit] is "inches".
		// Are there any cameras whose unit is "mm"?

		StringTokenizer t = new StringTokenizer(ccdxr);
		// decode the resolution string
		String numStr = t.nextToken("/");
		String denStr = stripFirstChar(t.nextToken(" "));
		String untStr = t.nextToken();

		double num = Double.parseDouble(numStr);
		double den = Double.parseDouble(denStr);
		double w;

		// convert from inches to mm, if needed.
		// this is experimental
		if (untStr.equals("mm")) {
			w = (double) (num / den * exifImageWidth(meta));
			// System.out.println("mm");
		} else {
			w = (double) (num / den * 25.4 * exifImageWidth(meta));
		}

		return w;
	}

	// 35mm focus helper method.
	// return the height of the CCD in mm.
	private double ccdHeight(Map meta) {
		String ccdyr = getValueForKey(meta, "Focal Plane Y Resolution");
		if (ccdyr == null || ccdyr.length() == 0)
			return 0.0;
		// the resolution is a string, containing a fraction
		// [num]/[den] [unit]
		// for Canon camera's, [unit] is "inches".
		// Are there any cameras whose unit is "mm"?

		StringTokenizer t = new StringTokenizer(ccdyr);
		// decode the resolution string
		String numStr = t.nextToken("/");
		String denStr = stripFirstChar(t.nextToken(" "));
		String untStr = t.nextToken();

		double num = Double.parseDouble(numStr);
		double den = Double.parseDouble(denStr);
		double h;

		// convert from inches to mm, if needed.
		// this is experimental
		if (untStr.equals("mm")) {
			h = (double) (num / den * exifImageHeight(meta));
			// System.out.println("mm");
		} else {
			h = (double) (num / den * 25.4 * exifImageHeight(meta));
		}

		return h;
	}

	/**
	 * Return the 35mm equivalent focal length.
	 * @param focalLength	The JAlbum focalLength parameter.
	 * @param cameraModel	The JAlbum cameraModel parameter.
	 * @param meta		The file's meta-information map.
	 * @return the 35mm equivalent focal length.
	 */
	public int focus35mm(String focalLength, String cameraModel, Map meta) {
		double f = numericPart(focalLength);
		// There are 2 ways of computing the 35mm equivalent focus:
		// 
		// 1. Read the conversion parameters from cameras.properties
		if (camera_info != null && cameraModel != null) {
			double multiplier = 0.0, offset = 0.0;
			if (camera_info.containsKey(cameraModel + ".focus_multiplier")) {
				multiplier = Double.parseDouble(camera_info.get(cameraModel + ".focus_multiplier").toString());
			}
			if (camera_info.containsKey(cameraModel + ".focus_offset")) {
				offset = Double.parseDouble(camera_info.get(cameraModel + ".focus_offset").toString());
			}
			return (int)(f * multiplier + offset + 0.5);
		}

		// 2. Use the Focal Plane X/Y resolution (if present)
		// Problem: if image is already rotated, the ccdWidth will return the
		//          actual ccdHeight and render the calculation invalid
		// Solution: take height if height > width
		// Problem: if the image was cropped, then the computation does
		//          NOT work.
		// Solution: disable this type of computation if your album
		//           contains cropped images

		if (allow_computation_with_focal_plane_resolution == true) {
			if (exifImageHeight(meta) > exifImageWidth(meta)) {
				if (ccdHeight(meta) > 0.001)
					return (int)(f * 35.0 / ccdHeight(meta) + 0.5);
			} else {
				if (ccdWidth(meta) > 0.001)
					return (int)(f * 35.0 / ccdWidth(meta) + 0.5);
			}
		}

		// Fallback:
		return 0;
	}

}
